#!/usr/bin/perl

# Copyright (C) 2005 John J. Chew, III <jjchew@math.utoronto.ca>
# All Rights Reserved

package TSH::Division::FindExtremeGames;

use strict;
use warnings;

=pod

=head1 NAME

TSH::Division::FindExtremeGames - algorithms for finding high/low wins/losses

=head1 SYNOPSIS

  TSH::Division::FindExtremeGames::Search($dp, $count, \&filter, \&ranker);

=head1 ABSTRACT

This module is used by the HighLoss, HighWin, LowLoss and LowWin
commands to generate a list of the top however many games, arbitrarily
selected and ranked.

Games are represented as lists: [$score1, $score2, $p1, $p2, $r0].

=cut

sub Search ($$&&);

=head1 DESCRIPTION

=over 4

=cut
 
=item $listp = Search($dp, $count, \&filter, \&ranker);

Return a reference to a list of up to $count games, each represented as
[$score1, $score2, $p1, $p2, $r0].  Only games which return true when
passed to &filter will be included, and if there are more than $count,
the list will be sorted using &ranker as a sortsub and only the first
$count members returned.  &ranker must be prototyped (see perldoc -f
sort).

=cut

sub Search ($$&&) {
  my $dp = shift;
  my $count = shift;
  my $filter = shift;
  my $ranker = shift;

  my @entries;
  for my $p ($dp->Players()) {
    for my $r0 (0..$p->CountScores()-1) {
      my $gamep 
        = [$p->Score($r0), $p->OpponentScore($r0), $p, $p->Opponent($r0), $r0];
      next unless &$filter($gamep);
      if (@entries < $count) {
	push(@entries, $gamep);
        }
      elsif (&$ranker($gamep, $entries[$count-1]) < 0) {
	push(@entries, $gamep);
	if (@entries > 2 * $count) {
	  @entries = sort $ranker @entries;
	  splice(@entries, $count); # not shared, so splice is thread-safe
	  }
        }
      }
    }
  @entries = sort $ranker @entries;
  if (@entries > $count) {
    splice(@entries, $count); # not shared, so splice is thread-safe
    }
  return \@entries;
  }


=head1 BUGS

None known.

=cut

1;
